"""Contain classes for defining logging and FastAPI"""
from typing import Dict, Any, TypedDict
from pathlib import Path
import logging
import json

import colorlog
from fastapi import FastAPI, Request, Depends
from fastapi.staticfiles import StaticFiles

from ArgosAnalysePlatform.api import router
from ArgosAnalysePlatform.database import DatabaseConnections
from ArgosAnalysePlatform.plugins import import_plugins
from ArgosAnalysePlatform.plugin_manager import plugmanager

global_dependencies = {}
SETTINGS_FILE = "settings.json"

logger = logging.getLogger(__name__)


class LoggingConfiguration:
    """Logging configuration for the base logger

    Establishes the handlers for the logger with the name "ArgosAnalysePlatform"
    """

    def __init__(self, level: str):
        """Initialize the base logger

        :param level: The logging level for the StreamHandler.
            Should be DEBUG, INFO, WARNING, CRITICAL or ERROR.
        """
        self.logger = logging.getLogger("ArgosAnalysePlatform")
        self.logger.setLevel("DEBUG")
        self.handler = colorlog.StreamHandler()
        self.handler.setLevel(level)
        self.handler.setFormatter(
            colorlog.ColoredFormatter(
                "%(log_color)" "s%(levelname)" "s:%(name)" "s:%(message)s"
            )
        )
        self.logger.addHandler(self.handler)


class ApiAppSettings(TypedDict):
    """Explicit types for the load_from_settings parameter for
    APIApp."""

    reflect: bool
    debug: bool
    external_plugins_folder: str


class APIApp(FastAPI):
    """Manage the FastAPI instance"""

    settings: Dict[str, Any]

    def __init__(
        self,
        load_from_settings: ApiAppSettings = None,
        fastapi_args=None,
        fastapi_kwargs=None,
    ):
        """Create the FastAPI app to be used by uvicorn.

        The app will include the router from ArgosAnalysePlatform.api in all
        connections. The router is added with a dependency to add the
        instantiated database as an attribute `request.state.db`.

        :param load_from_settings: A dict of settings to use.

        :return

        """
        super().__init__(*(fastapi_args or []), **(fastapi_kwargs or {}))

        if load_from_settings is None:  # pragma: no cover
            with open(SETTINGS_FILE, "r") as f:
                self.settings = json.load(f)
                logger.debug(f"Loading APIApp with settings {self.settings}")
        else:
            self.settings = load_from_settings

        # Initialize the logger
        LoggingConfiguration("DEBUG" if self.settings["debug"] else "INFO")

        # Load plugins
        if "external_plugins_folder" in self.settings:
            import_plugins(Path(self.settings["external_plugins_folder"]))

        import_plugins(Path(__file__).parent / "plugins", Path(__file__).parent.parent)

        # Define global FastAPI route dependencies
        def get_db(request: Request):  # pragma: no cover
            request.state.db = database

        if self.settings["use_database"]:
            database = DatabaseConnections.from_file(
                "connections.yaml"
            )

            if self.settings["reflect"]:
                database.reflect_database()
        else:
            database = []

        dependencies = [Depends(get_db)]
        global_dependencies["get_db"] = get_db

        self.include_router(router, dependencies=dependencies)
        self.include_router(plugmanager.router, dependencies=dependencies)
        self.mount(
            "/static",
            StaticFiles(directory=str((Path(__file__).parent / "html").absolute())),
            name="static",
        )
