"""Entrypoint for the entire system."""
import logging
import json

import uvicorn
import click

from ArgosAnalysePlatform import APIApp, SETTINGS_FILE


logger = logging.getLogger(__name__)


@click.command()
@click.option("--external-plugins", help="Path to external plugins")
@click.option("--debug", help="Path to external plugins", is_flag=True)
@click.option(
    "--no-reflect", help="Do not reflect the database tables if set.", is_flag=True
)
@click.option("--port", help="The port to run the server on", default=8000)
@click.option("--host", help="The host to listen on", default="127.0.0.1")
@click.option(
    "--disable-database",
    help="If set, disable connecting to the database",
    is_flag=True,
)
def main(external_plugins, debug, no_reflect, port, host, disable_database):
    """Run the Argos Analysis platform.

    Will launch the API with specified options.

    """
    settings = {}

    if external_plugins:
        settings["external_plugins_folder"] = external_plugins

    settings["reflect"] = not bool(no_reflect)
    settings["debug"] = debug
    settings["use_database"] = not disable_database

    if debug:
        # If running in debug mode, reload is set to True.
        # Reload will run the actual FastAPI instance in another Python
        # instance than this, which drops all variables/parameters that are set.
        # Counter this by storing the settings in a .json file, then
        # loading it at runtime.
        app = "ArgosAnalysePlatform.__main__:APIApp"
        with open(SETTINGS_FILE, "w") as f:
            json.dump(settings, f)
    else:
        app = APIApp(
            load_from_settings=settings,
        )

    uvicorn.run(
        app,
        host=host,
        port=port,
        factory=debug,
        reload=debug,
    )


if __name__ == "__main__":
    main(obj={})  # pylint: disable=unexpected-keyword-arg,no-value-for-parameter
