# pylint: disable=missing-class-docstring
"""Define the main API endpoints of the system."""
from datetime import datetime
from typing import Dict, Any, List, Union, Optional, Tuple
import xml.etree.ElementTree as ET

from dateutil.parser import isoparse
from fastapi import APIRouter, Request, Depends, Response, Query, Form
from pydantic import BaseModel  # pylint: disable=no-name-in-module

from ArgosAnalysePlatform.plugin_manager import plugmanager

router = APIRouter()


@router.get(
    "/plugins/categories",
    response_model=Dict[str, List[str]],
    summary="Get all categories",
)
def get_categories(_: Request):
    """Return the names of all registered targets in each category.

    Gets its data from :meth:`ArgosAnalysePlatform.plugin_manager.PluginManager.get_all_plugins`

    """
    return plugmanager.get_all_plugins()


# --- Search endpoint ---


class SearchModel(BaseModel):
    """Input model for the :func:`search` endpoint.

    More detailed documentation can be found in Swagger.
    """

    type: str = Query("", description="The type of data (timeseries, table, etc.)")
    target: str = Query("", description="Query field value")
    category: Optional[str] = Query("", description="The plugin category to use.")


@router.post("/search", summary="Search for categories")
def search(
    _: Request,
    data: SearchModel,
):
    """Return the names of all registered targets"""

    category = data.dict()["category"]
    return plugmanager.get_plugins(category)


# --- / endpoint ---


@router.get("/", summary="Test the connection to the server")
def test():
    """Connection check"""
    return {}


# --- /query endpoint ---


class TimeRangeType(BaseModel):
    """Dict of from and to time range.

    Used in :class:`JSONQueryInputType`


    More detailed documentation can be found in Swagger.
    """

    from_: str = Query(..., description="From time in iso format (YYYY-MM-dd HH:mm:ss)")
    to: str = Query(..., description="To time in iso format (YYYY-MM-dd HH:mm:ss)")

    class Config:
        fields = {"from_": "from"}


class ScopedVarsModel(BaseModel):
    """Dict of scoped vars.

    Used in :class:`JSONQueryInputType`

    More detailed documentation can be found in Swagger.
    """

    value: Union[List[str], str, int] = Query(
        ..., description="The value of the scoped variable."
    )


class TargetsModel(BaseModel):
    """Model for describing targets (=the plugin names)

    Used in :class:`JSONQueryInputType`

    More detailed documentation can be found in Swagger.
    """

    target: str = Query(..., description="Target plugin name")
    data: Any = Query(..., description="Extra data for the plugin.")


class JSONQueryInputType(BaseModel):
    """Final model of the input type for :func:`query`.

    More detailed documentation can be found in Swagger.
    """

    range: TimeRangeType
    intervalMs: int = Query(..., description="Time separation interval in ms.")
    scopedVars: Dict[str, ScopedVarsModel] = Query(
        ..., description="Variables defined in the dashboard."
    )
    targets: List[TargetsModel] = Query(..., description="Target plugins.")


class JSONQueryTimeseriesResponseModel(BaseModel):
    """Response model for timeseries-like responses from :func:`query`.

    The other alternative is :class:`JSONQueryTableResponseModel`

    More detailed documentation can be found in Swagger.
    """

    target: str = Query(..., description="The name of the dataset.")
    datapoints: List[Tuple[int, datetime]] = Query(
        ..., description="List of datapoints for the dataset."
    )


class JSONQueryTableColumn(BaseModel):
    """Define the columns for :class:`JSONQueryTableResponseModel`

    More detailed documentation can be found in Swagger.
    """

    text: str = Query(..., description="The name of the column")
    type: str = Query(..., description="Datatype of the column")


class JSONQueryTableResponseModel(BaseModel):
    """Response model for table-like responses from :func:`query`.

    The other alternative is :class:`JSONQueryTimeseriesResponseModel`

    More detailed documentation can be found in Swagger.
    """

    columns: List[JSONQueryTableColumn]
    rows: List[Any] = Query(
        ...,
        description="Rows of the table. Must have as many elements as count of columns.",
    )
    type: str = "table"


@router.post(
    "/query",
    response_model=Union[
        List[JSONQueryTimeseriesResponseModel], List[JSONQueryTableResponseModel]
    ],
)
def query(request: Request, inp: JSONQueryInputType):
    """Query the database based on the supplied plugins

    Query the database for data based on the selected target.
    Target must be defined in plugins with\
    :meth:`ArgosAnalysePlatform.plugin_manager.PluginManager.register_plugin`
    """
    inp = inp.dict()

    start_time = isoparse(inp["range"]["from_"])
    end_time = isoparse(inp["range"]["to"])
    resolution = inp["intervalMs"] // 1000 // 60

    scanners = inp["scopedVars"]["scannere"]["value"]
    metrics = [(t["target"], t["data"]) for t in inp["targets"]]
    response = []
    for metric, data in metrics:
        for name, session in request.state.db.sessioniter(scanners):

            response.extend(
                plugmanager.call_plugin(metric, "")(
                    session,
                    name,
                    start_time,
                    end_time,
                    resolution,
                    extra_data=data,
                )
            )

    return response


# --- xml/query endpoint ---


async def parse_async_form(request: Request):  # pytest: no cover
    """Convert FastAPI async form to normal dict.

    FastAPI passes form parameters as async objects. These need to be converted
    to be used synchronously.
    """
    return dict(await request.form())


class XMLQueryModel(BaseModel):
    """Model for accepting inputs to :func:`xml_query`"""

    start_time: str
    end_time: str
    actions: str
    targets: str
    resolution: str

    @classmethod
    def as_form(
        cls,
        start_time: str = Form(
            "", description="Start time in isoformat (YYYY-MM-dd HH:mm:ss)"
        ),
        end_time: str = Form(
            "", description="End time in isoformat (YYYY-MM-dd HH:mm:ss)"
        ),
        actions: str = Form(
            "", description="Which plugin(s) to call as comma separated list"
        ),
        targets: str = Form(
            "", description="Which scanners to run on as comma separated list"
        ),
        resolution: str = Form("", description="Time separation interval in ms"),
    ) -> "XMLQueryModel":
        """Tell FastAPI to expect the data as HTML Form instead of JSON."""
        return cls(
            start_time=start_time,
            end_time=end_time,
            actions=actions,
            targets=targets,
            resolution=resolution,
        )


class XMLRow(BaseModel):
    """Model for describing :class:`XMLResponseModel` rows."""

    column: Dict[str, Any]


class XMLResponseModel(BaseModel):
    """Model for describing the return type of :func:`xml_query`."""

    row: Dict[str, XMLRow]


xml_query_responses = {
    200: {
        "content": {
            "application/xml": {
                "schema": {**XMLResponseModel.schema(), **{"xml": {"name": "root"}}}
            }
        }
    }
}


@router.post("/xml/query", responses=xml_query_responses)
def xml_query(
    request: Request,
    required_data: XMLQueryModel = Depends(XMLQueryModel.as_form),
    all_data: dict = Depends(parse_async_form),
):
    """Query data from the database based on plugins.

    Will not reject any unknown parameters, but must include a list of
    required parameters.

    :param required_data: Fields that are required. See parameters of\
    :func:`XMLQueryModel.as_form`.
    :param all_data: All supplied fields, including required data.

    """

    required_data = required_data.dict()

    actions = required_data["actions"].split(",")
    targets = required_data["targets"].split(",")

    start_time = required_data.get("start_time")
    end_time = required_data.get("end_time")

    if start_time:
        start_time = isoparse(start_time)
    if end_time:
        end_time = isoparse(end_time)

    resolution = required_data.get("resolution")
    if resolution:
        resolution = int(resolution) // 1000 // 60

    root = ET.Element("root")
    for action in actions:
        for name, session in request.state.db.sessioniter(targets):
            result = plugmanager.call_plugin(action, "report_builder")(
                session,
                name,
                start_time,
                end_time,
                resolution,
                extra_data=all_data,
            )

            if isinstance(result, list):
                root.extend(result)
            elif isinstance(result, ET.Element):
                root.append(result)
            else:
                raise ValueError(
                    f"Return type of XML plugin must be Element or List, got {type(result)}"
                )

    return Response(content=ET.tostring(root), media_type="application/xml")
