"""Handle connections to the an arbitrary amount of databases. """
import sys
from pathlib import Path
from typing import List, Union, Iterable, Tuple
import logging

from sqlalchemy import create_engine, MetaData
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker, Session
import yaml

logger = logging.getLogger(__name__)


class DatabaseConnections:
    """Organize connections to the different databases.

    Currently only supports connections through ODBC.

    Which driver is used depends on operating system.

        - Windows: SQL Server
        - Everything else: ODBC Driver 17 for SQL Server.
    """

    if sys.platform == "win32":
        driver = "SQL+Server"
    else:
        driver = "ODBC+Driver+17+for+SQL+Server"
    engine_string_template = (
        f"mssql+pyodbc://{{username}}:{{password}}@{{ip}}/{{tablename}}?driver={driver}"
    )
    connections: dict
    tables: dict

    @classmethod
    def from_file(cls, path: Union[str, Path]):  # pragma: no cover
        """

        :param path: Path to the file to connect to."""
        with Path(path).open("r") as connections_file:
            connections_dict = yaml.safe_load(connections_file)

        return cls(connections_dict)

    def __init__(self, dct: dict):
        """
        :param dct: Dict of configuration data
        """

        self.connections = {}
        # Get the shared dict if it exists.
        shared_settings = dct.get("shared", {})

        for name, details in dct.get("connections", {}).items():
            engine = create_engine(
                DatabaseConnections.engine_string_template.format(
                    # Merge the details and shared_settings dict, then
                    # expand the result as kw-arguments.
                    **{**details, **shared_settings}
                )
            )
            self.connections[name] = {
                "engine": engine,
                "sessionmaker": sessionmaker(bind=engine),
            }

        if len(self.connections) == 0:
            logger.warning("Connections file contained no connections.")

    def reflect_database(self) -> None:  # pragma: no cover
        """Reflect database tables from the given engine"""
        # Read the database tables and _reflect_ them into local
        # sqlalchemy objects that can be used for querying.
        if len(self.connections) > 0:
            self.meta = MetaData()
            self.meta.reflect(bind=list(self.connections.values())[0]["engine"])
            self.__class__.tables = self.meta.tables
        else:
            logger.error("Could not reflect on empty connections")

    def sessioniter(
        self, conns: Union[str, List[str]] = None
    ) -> Iterable[Tuple[str, Session]]:
        """Get sessions from the speicified connections.

        Get an iter of the chosen sessions.

        :param conns: The chosen session(s).
        :return: An iter of the scanner name and Session.
        """
        conns = conns if isinstance(conns, list) else [conns]
        yield from (
            (k, c["sessionmaker"]()) for k, c in self.connections.items() if k in conns
        )
