"""Plugin manager for the API and static endpoints. """
import collections
import functools
import traceback
import logging
from typing import Dict, List, Callable

from fastapi import APIRouter, HTTPException


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class PluginManager:
    """Manage plugins for the /query endpoint."""

    _plugins = collections.defaultdict(lambda: {})

    def __init__(self):
        self.router = APIRouter()

    def api_post(self, endpoint_name: str, *args, **kwargs) -> Callable:
        """Add a POST endpoint to the API.

        :param endpoint_name: URI for the endpoint
        :param args: Other arguments to FastAPI
        :param kwargs: Other arguments to FastAPI
        """
        logger.debug(f"Adding POST endpoint {endpoint_name}")
        return self.router.post(endpoint_name, *args, **kwargs)

    def api_get(self, endpoint_name: str, *args, **kwargs) -> Callable:
        """Add a GET endpoint to the API.

        :param endpoint_name: URI for the endpoint
        :param args: Other arguments to FastAPI
        :param kwargs: Other arguments to FastAPI
        """
        logger.debug(f"Adding GET endpoint {endpoint_name}")
        return self.router.get(endpoint_name, *args, **kwargs)

    def register_plugin(self, name: str, category: str = ""):
        """Decorator factory to register plugins

        Registered plugins will appear in /search with the supplied name.
        The decorated function can then be run with\
         :func:`ArgosAnalysePlatform.plugin_manager.PluginManager.call_plugin`.

        :param name: The name of the plugin to be registered.
        :param category: The category to register the plugin in.
        """

        def wrapper(func):
            self._plugins[category][name] = func
            logger.debug(f"Registering plugin {name} with category='{category}'")

            @functools.wraps(func)
            def wrapped(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapped

        return wrapper

    def get_plugins(self, category: str) -> List[str]:
        """Return the names of all plugins in the specified category.

        :param category: Only show plugins in this category.
        """
        return list(self._plugins[category].keys())

    def get_all_plugins(self) -> Dict[str, List[str]]:
        """Return the name of all registered plugins"""
        return {k: list(v.keys()) for k, v in self._plugins.items()}

    def call_plugin(self, plugin_name: str, category: str) -> Callable:
        """Call a registered plugin

        :param plugin_name: The name of the plugin
        :param category: The category/namespace of the plugin.

        :return: A callable to the plugin. Wrapped in error handling logic for API use.
        """

        plugin_callable = self._get_plugin(plugin_name, category)

        @functools.wraps(plugin_callable)
        def run(*args, **kwargs):
            try:
                return plugin_callable(*args, **kwargs)
            except HTTPException as err:
                err.detail = {"msg": err.detail, "plugin": plugin_name}
                raise
            except Exception as err:
                traceback.print_exc()

                raise HTTPException(
                    status_code=500,
                    detail={
                        "type": str(type(err).__name__),
                        "msg": str(err),
                        "plugin": plugin_name,
                    },
                ) from err

        return run

    def _get_plugin(self, plugin_name: str, category: str = "") -> Callable:
        """Find a plugin and wrap it in a 404 if it does not exsist."""
        try:
            return self._plugins[category][plugin_name]
        except KeyError as err:
            raise HTTPException(
                status_code=404,
                detail=f"Could not find plugin {plugin_name} "
                f"in category {category}",
            ) from err


plugmanager = PluginManager()
