# pylint: disable=missing-class-docstring
"""Simulating changes in production data based on limits."""
import collections
import logging
from datetime import datetime
from itertools import groupby
from typing import Union, Dict, Any, List, Tuple
import statistics

from fastapi import Request, Query
import matplotlib.pyplot as plt
from matplotlib.ticker import PercentFormatter, MultipleLocator
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from sqlalchemy import func, and_, case, distinct
from sqlalchemy.orm import Session
from sqlalchemy.engine.result import RowProxy

from ArgosAnalysePlatform import plugmanager, DatabaseConnections
from ArgosAnalysePlatform.plugins.regfile_reader import ArgosRegistry


logger = logging.getLogger("ArgosAnalysePlatform.plugin_manager.analysis")
logger.setLevel(logging.DEBUG)

BaseModel.__init__.__doc__ = ""


def _generate_cumulative_percentage(y):

    cumulative_percentage = [yy / sum(y) * 100 for yy in y]

    return [
        sum(cumulative_percentage[: i + 1]) for i in range(len(cumulative_percentage))
    ]


def generate_pareto_plot(x: List[int], y: List[int]) -> plt.Figure:
    """Generate a pareto plot from the supplied data

    :param x: The x axis
    :param y: The y axis

    :return: The figure of the plotted data.
    """

    fig, ax = plt.subplots()

    ax.bar(x, y)

    # Get prettier axis ticks.
    ax.xaxis.set_major_locator(MultipleLocator(1))
    ax.yaxis.set_major_locator(MultipleLocator(1))

    # Assemble a line graph of cumulative boards.
    cumulative_y = _generate_cumulative_percentage(y)

    tot_ax = ax.twinx()
    tot_ax.plot(x, cumulative_y, marker="D", ms=7, color="C1")
    tot_ax.yaxis.set_major_formatter(PercentFormatter())
    ax.yaxis.set_major_locator(MultipleLocator(1))
    tot_ax.tick_params(axis="y", color="C1")

    return fig


def _query_pareto(
    session: Session,
    where_argosnames: List[str],
    start_time: datetime,
    end_time: datetime,
) -> Dict[Any, List[Tuple[Any, int]]]:
    """Query required data to make pareto charts.

    Returns a list of argos names (filtered by parameter), amount of boards
    grouped by the am

    :param session: Database session
    :param where_argosnames: A list of the argosnames to filter on.

    :return: Dict of dict of the results, grouped first by class, then type/defect count

    """
    Defects = DatabaseConnections.tables["Defects"]
    Boards = DatabaseConnections.tables["Boards"]

    # Get all boards that has the selected where_argosnames as worst defect.
    worst_subq = (
        session.query(Defects.c.BoardID, Defects.c.ArgosName)
        .join(Boards)
        .filter(Boards.c.Time.between(start_time, end_time))
        .filter(Defects.c.IsWorstDefect == 1)
        .filter(Defects.c.ArgosName.in_(where_argosnames))
        .subquery()
    )

    old_class = func.max(Defects.c.Class)
    # Get the highest class that is not worst defect.
    new_class = func.max(
        case(
            [(Defects.c.ArgosName != worst_subq.c.ArgosName, Defects.c.Class)], else_=0
        )
    )

    filter_values = (
        session.query(
            Defects.c.BoardID,
            old_class.label("old_class"),
            new_class.label("new_class"),
        )
        .join(
            worst_subq,
            worst_subq.c.BoardID == Defects.c.BoardID,
        )
        .group_by(Defects.c.BoardID)
        # Only return boards that actually change class if worst defect is
        # excluded.
        .having(new_class < old_class)
        .subquery()
    )

    query = (
        session.query(
            Defects.c.BoardID.label("bid"),
            Defects.c.ArgosName.label("type"),
            func.count(Defects.c.ID).label("defect_count"),
            filter_values.c.old_class,
            filter_values.c.new_class,
        )
        .join(
            # Only include boards that are part of where_argosnames and that
            # has them as worst.
            worst_subq,
            and_(
                worst_subq.c.BoardID == Defects.c.BoardID,
                worst_subq.c.ArgosName == Defects.c.ArgosName,
            ),
        )
        # Exclude boards that would not be affected by the change.
        .join(filter_values, filter_values.c.BoardID == Defects.c.BoardID)
        .group_by(
            Defects.c.BoardID,
            Defects.c.ArgosName,
            filter_values.c.old_class,
            filter_values.c.new_class,
        )
        # .subquery()
        .order_by(new_class, Defects.c.ArgosName, func.count(Defects.c.ID))
        .all()
    )

    groupby_new_class = {}

    # Group the result by new class
    for k, v in groupby(
        sorted(query, key=lambda x: x.new_class), key=lambda x: x.new_class
    ):
        # Group the result by its type and defect count
        groupby_type_defectcount = [
            (*k2, len(list(v2)))
            for k2, v2 in groupby(v, key=lambda x: (x.type, x.defect_count))
        ]

        groupby_new_class[k] = groupby_type_defectcount

    # Dict of the reults grouped first by class, then by type/defect count.
    return groupby_new_class


def _fill_partial_axis(data: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
    """Fill missing values in a dataset of x and y values."""

    # Get X axis by transposing the defect_count/board_count data
    x, _ = zip(*data)
    # Find and insert missing x values.
    missing = set(range(1, max(x))) - set(x)
    data.extend((n, 0) for n in missing)

    return sorted(data, key=lambda x: x[0])


def _generate_pareto_xy_from_db(
    session: Session,
    defect_argosnames: List[str],
    start_time: datetime,
    end_time: datetime,
) -> Dict[str, Dict[int, Tuple[List[int], List[int]]]]:
    """Generate x and y axis data from the returned database rows

    :param session: Database session
    :param defect_argosnames: The defect names to simulate.

    :return: X and Y axis of the data, grouped by ArgosName, then defect class.
    """
    query_data = _query_pareto(session, defect_argosnames, start_time, end_time)

    pareto_result = collections.defaultdict(dict)

    for cls, values in query_data.items():
        all_defects = collections.defaultdict(list)
        # Extract ArgosName from the board results.
        for argos_name, defect_count, board_count in values:
            all_defects[argos_name].append((defect_count, board_count))

        for argos_name, data in all_defects.items():

            # Get X and Y axies by transposing list of (x,y) coordinates.
            x, y = zip(*_fill_partial_axis(data))
            pareto_result[argos_name][cls] = (x, y)

    return pareto_result


def find_most_valuable_change(
    y: List[int], current: int, threshold: int = 1, max_diff: int = 4
):
    """Based on the Y values, find the most valuable difference.

    y must contain one entry per x axis.

    Computes most valuable change by iteratively computing the standard deviation
    in y values. When this value is above the threshold, return the current position.

    If the standard deviation never crosses the threshold, return the last x position.


    :param y: The y values to base the change on.
    :param current: The current x value.
    :param max_diff: The maximum difference between current and new value.

    :return: The X position for the most valuable change.
    """

    if len(y) < 2:
        return current

    for i in range(current + 2, len(y) + 1):
        stdev = statistics.stdev(y[current:i])

        if stdev > threshold:
            return i - 1

        if i - current >= max_diff:
            return i

    return len(y)


anyjson = Union[List, Dict[str, Any]]


def _get_argosnames_from_subtypes(
    session: Session, subtypes: List[str]
) -> Dict[str, str]:
    """Get all argosnames that belong to the selected subtypes.

    :param session: The session to get data from.
    :param subtypes: List of subtypes to convert to ArgosName.

    :return: List of ArgosNames.
    """
    Defects = DatabaseConnections.tables["Defects"]

    data = (
        session.query(distinct(Defects.c.SubType), Defects.c.ArgosName)
        .filter(Defects.c.SubType.in_(subtypes))
        .all()
    )

    result = {k: list(v2[1] for v2 in v) for k, v in groupby(data, key=lambda x: x[0])}

    return result


class GetArgosnamesModel(BaseModel):
    """Specify subtypes for :func:`manual_simulate_rows`"""

    subtypes: List[str]
    old_limits: Dict[str, Dict[str, str]] = None
    scanners: List[str] = None


@plugmanager.api_post("/manual_simulate_rows")
def manual_simulate_rows(request: Request, data: GetArgosnamesModel):
    """
    Takes a list of subtypes. For each subtype, find out how many boards' class may
    improve if the specified defect subtypes are ignored.

    See a pareto plot of these boards through the URL "/static/<argos name><defect class>".
    """
    old_limit = ArgosRegistry.from_dct(data.old_limits) or ArgosRegistry.from_file(
        "argos.reg", "PRS.SwansonPRS.200120"
    )

    for _, session in request.state.db.sessioniter(data.scanners or ["scanner1"]):
        result = {}
        for subtype, argos_names in _get_argosnames_from_subtypes(
            session, data.subtypes
        ).items():
            result[subtype] = {a_name: old_limit.get(a_name) for a_name in argos_names}

        return result


class BoardsByClassModel(BaseModel):
    """Model for :func:`boards_by_class`"""

    start_time: datetime
    end_time: datetime
    scanners: List[str]


@plugmanager.api_post("/boards_by_class")
def boards_by_class(request: Request, data: BoardsByClassModel):
    """Get count of boards per class"""
    for _, session in request.state.db.sessioniter(data.scanners):
        return _get_boards_by_class(session, data.start_time, data.end_time)


def _get_boards_by_class(session, start_time, end_time):
    """Query data for count of boards per class."""
    Boards = DatabaseConnections.tables["Boards"]
    data = (
        session.query(Boards.c.Class, func.count(Boards.c.BoardID).label("count"))
        .filter(Boards.c.Time.between(start_time, end_time))
        .group_by(Boards.c.Class, Boards.c.ClassName)
        .all()
    )

    sum_boards = sum(v.count for v in data)
    result = {v.Class: [v.count, round(v.count / sum_boards * 100, 2)] for v in data}

    return result


# pylint: disable=too-many-locals
# This function has a lot of locals, but after consideration its probably
# good enough to leave it as one function instead of splitting it up for
# pylints sake.
def simulate_production_change(
    session,
    argos_names,
    start_time,
    end_time,
    *,
    new_limits: Dict[str, Dict[str, int]] = None,
    suggest_limit_callable: callable = find_most_valuable_change,
    scanner_name="",
    old_limits: Dict[str, Dict[str, str]] = None,
):
    """Simulate change in production output

    Simulate changes in production output if the given argos names had different
    limits.

    :param session: The session to get from
    :param argos_names: The argos name to simulate on.
    :param start_time: From datetime
    :param end_time: End datetime
    :param new_limits: If given, use these limits. If not given, suggest
        new ones based on suggest_limit_callable.
    :param suggest_limit_callable: Callable to determine which limit to suggest.
        Must take a list of y values and the current limit. Must return the new
        limit as a single integer.


    """
    result = collections.defaultdict(list)

    old_limit = ArgosRegistry.from_dct(old_limits) or ArgosRegistry.from_file(
        "argos.reg", "PRS.SwansonPRS.200120"
    )
    boards_group_class = _get_boards_by_class(session, start_time, end_time)

    pareto_data = _generate_pareto_xy_from_db(
        session, argos_names, start_time, end_time
    )

    for argos_name, data in pareto_data.items():
        for defect_class, xy in data.items():
            print(f"Generating plot for {argos_name}")
            fig = generate_pareto_plot(*xy)
            fig.savefig(f"ArgosAnalysePlatform/html/{argos_name}{defect_class}.png")

            if new_limits:
                new_limit = new_limits.get(argos_name, {}).get(
                    f"{defect_class+1}_{defect_class+2}"
                )
            else:
                new_limit = suggest_limit_callable(
                    data[defect_class][1],
                    old_limit.get_upper_limit(argos_name, defect_class),
                )

            delta = sum(
                data[defect_class][1][
                    old_limit.get_upper_limit(argos_name, defect_class) : new_limit
                ]
            )

            percent_increase = round(
                delta / (boards_group_class[defect_class][0] + delta) * 100, 2
            )

            result[argos_name].append(
                (
                    defect_class,
                    old_limit.get_upper_limit(argos_name, defect_class),
                    new_limit,
                    # Increase in production from old to new limit.
                    delta,
                    percent_increase,
                    scanner_name,
                )
            )

            logger.debug(
                f"Most valuable change is from "
                f"{old_limit.get_upper_limit(argos_name, defect_class)} to "
                f"{new_limit}"
            )

    return result


class BaseSimulateModel(BaseModel):
    """Base model for data to :func:`simulate_images`

    Inherited by :class:`SuggestLimitsModel` and :class:`ManualLimitsModel`"""

    start_time: datetime = ""
    end_time: datetime = ""
    scanners: List[str]
    old_limits: Dict[str, Dict[str, str]] = None


class SuggestLimitsModel(BaseSimulateModel):
    subtypes: List[str]


class ManualLimitsModel(BaseSimulateModel):
    """Model for :func:`simulate_images_manual`"""

    limits: Dict[str, Dict[str, int]]


@plugmanager.api_post("/simulate_images")
def simulate_images(request: Request, data: SuggestLimitsModel):
    """Endpoint for suggesting simulation data.

    :param request: FastAPI request object
    :param defect_argosnames: The defect names to do the simulation on.
    """

    result = {}

    for scanner_name, session in request.state.db.sessioniter(data.scanners):
        argos_names = list(
            zip(*_get_argosnames_from_subtypes(session, data.subtypes).values())
        )[0]

        result = simulate_production_change(
            session,
            argos_names,
            data.start_time,
            data.end_time,
            scanner_name=scanner_name,
            old_limits=data.old_limits,
        )

    return result


@plugmanager.api_post("/simulate_images/manual")
def simulate_images_manual(request: Request, data: ManualLimitsModel):
    """Simulate production change based on manually specified new limits."""
    result = {}

    for scanner_name, session in request.state.db.sessioniter("scanner1"):
        result = simulate_production_change(
            session,
            list(data.limits.keys()),
            data.start_time,
            data.end_time,
            new_limits=data.dict()["limits"],
            scanner_name=scanner_name,
        )

    return result


def _get_types_and_subtypes(session: Session) -> List[RowProxy]:  # pragma: no cover
    """Get all types and subtypes from the database

    :param session: The session to get data from.
    """
    Defects = DatabaseConnections.tables["Defects"]

    group_order_by = (Defects.c.Type, Defects.c.SubType)
    data = (
        session.query(*group_order_by)
        .group_by(*group_order_by)
        .order_by(*group_order_by)
    )

    return data


class AvailableTypesChildrenResponse(BaseModel):
    """Model for specifying children in :class:`AvailableTypesResponse`"""

    text: str = Query(..., description="Name of the subtype")
    id: str = Query(..., description="Unique ID")


class AvailableTypesResponse(BaseModel):
    """Model for available types, returned by :func:`available_types`"""

    text: str = Query(..., description="Name of the defect type")
    children: List[AvailableTypesChildrenResponse]


class AvailableTypesModel(BaseModel):
    scanners: List[str]


@plugmanager.api_get("/available_types", response_model=List[AvailableTypesResponse])
def available_types(
    request: Request, scanners: AvailableTypesModel = None
) -> List[AvailableTypesResponse]:
    """Return available defect types and subtypes.

    Result is structured as a list of the defect types, with inner lists for the
    subtypes.
    """
    if scanners:
        scanners = scanners.scanners
    else:
        scanners = ["scanner1"]

    for _, session in request.state.db.sessioniter(scanners):
        result = []
        data = groupby(_get_types_and_subtypes(session), lambda x: x.Type)

        for defect_type, children in data:
            children_dicts = [
                {"id": child.SubType, "text": child.SubType} for child in children
            ]
            children_dicts.insert(
                0,
                {
                    "id": f"{'<**>'.join(c['id'] for c in children_dicts)}",
                    "text": f"All {defect_type}",
                },
            )

            result.append({"text": defect_type, "children": children_dicts})

        return result


@plugmanager.api_get("/available_scanners")
def avaliable_scanners(request: Request) -> List[Dict[str, str]]:
    """Get registered scanners."""
    return [{"id": v, "text": v} for v in request.state.db.connections.keys()]
