"""Plugins regarding Board related queries."""
from datetime import timedelta

from sqlalchemy import (
    func,
    text,
    distinct,
    desc,
    case,
    cast,
    DECIMAL,
    literal,
    and_,
)

from ArgosAnalysePlatform import plugmanager
from ArgosAnalysePlatform.database import DatabaseConnections
from ArgosAnalysePlatform.util import groupby


def group_by_time(session, resolution):
    """Groups result from queries in time-groups

    Parameters:
        session       (Session): Creates an instance for the program to work with the database
        resolution    (int): control size of time-groups

    Returns:
         Length       (Column): Length from Boards
         Class        (Column): Class from Boards
         ID           (Column): ID from Boards
         time         (Dateadd): Dateadd-object

    "DatabaseConnections.tables[]" creates a conncetion to a specified table.

    """
    Boards = DatabaseConnections.tables["Boards"]
    time = func.dateadd(
        text("minute"),
        resolution
        + func.datediff(text("minute"), 0, DatabaseConnections.tables["Boards"].c.Time)
        / resolution
        * resolution,
        0,
    ).label("time")

    return session.query(
        Boards.c.Length,
        Boards.c.Class,
        Boards.c.ID,
        time,
    ).subquery()


def get_boards_count(session, ___, start_time, end_time, resolution, *_, **__):
    """Counts number of boards produced

    Parameters:
        session       (Session): Creates an instance for the program to work with the database
        start_time    (datetime): start/from where the query will fetch data
        end_time      (datetime): end/to where the query will fetch data
        resolution    (int): control size of time-groups
        *_            declaring extra variables if necessary
        **__          declaring extra variables if necessary

    Returns:
         Length       (Column): Length from Boards
         Class        (Column): Class from Boards
         ID           (Column): ID from Boards
         time         (Dateadd): Dateadd-object
    """
    time_subq = group_by_time(session, resolution)
    data = (
        session.query(func.count(time_subq.c.ID), time_subq.c.time)
        .group_by(time_subq.c.time)
        .filter(time_subq.c.time.between(start_time, end_time))
        .order_by(time_subq.c.time)
        .all()
    )
    return data


@plugmanager.register_plugin("Boards produced")
def boards_count(session, metric, start_time, end_time, resolution, *_, **__):
    """Formats data from get_boards_count and sends to Grafana """
    data = get_boards_count(session, metric, start_time, end_time, resolution, *_, **__)
    result = [(d, int(((t + timedelta(hours=0)).timestamp()) * 1000)) for d, t in data]
    return [{"target": f"{metric} - Boards", "datapoints": result}]


def _query_number_of_boards_data(session, start_time, end_time, resolution):
    Boards = DatabaseConnections.tables["Boards"]
    time = func.dateadd(
        text("minute"),
        resolution
        + func.datediff(text("minute"), 0, Boards.c.Time) / resolution * resolution,
        0,
    ).label("time")

    t_sq = session.query(time).subquery()
    times = (
        session.query(t_sq.c.time)
        .group_by(t_sq.c.time)
        .order_by(t_sq.c.time)
        .subquery()
    )
    classes = session.query(distinct(Boards.c.Class).label("cls")).subquery()

    data = (
        session.query(classes.c.cls, func.count(Boards.c.ID), times.c.time)
        .select_from(classes)
        .join(times, literal(1) == literal(1))  # cross join
        .outerjoin(Boards, and_(Boards.c.Class == classes.c.cls, time == times.c.time))
        .filter(times.c.time.between(start_time, end_time))
        .group_by(times.c.time, classes.c.cls)
        .order_by(times.c.time, classes.c.cls)
    )

    return data


@plugmanager.register_plugin("Number of Repair, Reject, Good and Total in interval")
def number_of_boards_in_class(session, ___, start_time, end_time, resolution, *_, **__):
    """Returns number of boards produced in each class, including total"""

    result = []

    data = _query_number_of_boards_data(session, start_time, end_time, resolution).all()

    # sort by class, 'x[0]' refers to the position in the array which the function should sort by
    temp = groupby(data, key=lambda x: x[0])

    for k, v in temp:
        _, ant, t = zip(*v)
        data_points = list(zip(ant, t))
        result.append({"target": k, "datapoints": data_points})

    return result


def get_repair_percentage(session, ___, start_time, end_time, resolution, *_, **__):
    """Returns percentage of repair for each time-group in interval"""
    subq = group_by_time(session, resolution)

    data = (
        session.query(
            (
                cast(func.sum(case([(subq.c.Class == "1", 1)], else_=0)), DECIMAL(9, 2))
                / func.count(subq.c.ID)
            )
            * 100,
            subq.c.time,
        )
        .filter(subq.c.time.between(start_time, end_time))
        .group_by(subq.c.time)
        .order_by(subq.c.time)
        .all()
    )
    return data


@plugmanager.register_plugin("Boards classified Repair in percentage")
def boards_classified_percentage(
    session, metric, start_time, end_time, resolution, *_, **__
):
    """Get data from "get_repair_percentage" and returns in Grafana-friendly format"""
    data = get_repair_percentage(
        session, metric, start_time, end_time, resolution, *_, **__
    )

    return [{"target": "Metric", "datapoints": data}]


def get_line_speed(session, ___, start_time, end_time, resolution, *_, **__):
    """Get linespeed in meter/second

    Line speed is calulateed by: (Board-length(mm) / 1000) / (resolution(min) * 60 ) = m/s
    """
    subq = group_by_time(session, resolution)

    data = (
        session.query((func.sum(subq.c.Length) / 1000) / (resolution * 60), subq.c.time)
        .filter(subq.c.time.between(start_time, end_time))
        .group_by(subq.c.time)
        .order_by(subq.c.time)
        .all()
    )
    return data


@plugmanager.register_plugin("Line speed")
def line_speed(session, metric, start_time, end_time, resolution, *_, **__):
    """Get data from "get_line_speed" and returns in Grafana-friendly format"""
    data = get_line_speed(session, metric, start_time, end_time, resolution, *_, **__)
    return [{"target": "Linespeed", "datapoints": data}]


def get_product_info(session, _, start_time, end_time, __, *___, **____):
    """Fetches data(Series, Product, Customer, Time) about product"""
    Boards = DatabaseConnections.tables["Boards"]
    data = (
        session.query(
            Boards.c.Series, Boards.c.Product, Boards.c.Customer, Boards.c.Time
        )
        .filter(Boards.c.Time.between(start_time, end_time))
        .group_by(Boards.c.Series, Boards.c.Product, Boards.c.Customer, Boards.c.Time)
        .order_by(desc(Boards.c.Time))
        .all()
    )
    return data


@plugmanager.register_plugin("Product information")
def product_info(session, metric, start_time, end_time, resolution, *_, **__):
    """Fetchees queried data and returns data in Grafana-friendly format"""
    data = get_product_info(session, metric, start_time, end_time, resolution, *_, **__)
    return [
        {
            "columns": [
                {"text": "Series", "type": "String"},
                {"text": "Product", "type": "String"},
                {"text": "Customer", "type": "String"},
                {"text": "Time", "type": "time"},
            ],
            "rows": data,
            "type": "table",
        }
    ]
