"""Plugin for all Defect related queries."""
from itertools import groupby

from sqlalchemy import func, desc

from ArgosAnalysePlatform import plugmanager
from ArgosAnalysePlatform.database import DatabaseConnections
from ArgosAnalysePlatform.plugins.boards import group_by_time


def get_top_3_defects(session, ___, start_time, end_time, resolution, *_, **__):
    """Fetches top 3 defects which caused downgrade"""
    Defects = DatabaseConnections.tables["Defects"]
    subq = group_by_time(session, resolution)

    data = (
        session.query(Defects.c.Type, func.count(subq.c.ID))
        .filter(
            subq.c.time.between(start_time, end_time),
            Defects.c.IsWorstDefect == "1",
            subq.c.ID == Defects.c.BoardID,
        )
        .group_by(Defects.c.Type)
        .order_by(desc(func.count(subq.c.ID)))
        .limit(3)
        .all()
    )
    return data


@plugmanager.register_plugin("Top 3 Defect in time range")
def top_3_defects(session, metric, start_time, end_time, resolution, *_, **__):
    """Fetches data from "get_top_3_defects" and return in Grafana-friendly format"""
    data = get_top_3_defects(
        session, metric, start_time, end_time, resolution, *_, **__
    )
    return [
        {
            "columns": [
                {"text": "Type", "type": "string"},
                {"text": "Antall", "type": "number"},
            ],
            "rows": data,
            "type": "table",
        }
    ]


def get_repairs_by_defect_type(
    session, ___, start_time, end_time, resolution, *_, **__
):
    """Fetches repair count by type of defect"""
    subq = group_by_time(session, resolution)
    Defects = DatabaseConnections.tables["Defects"]

    data = (
        session.query(Defects.c.Type, func.count(Defects.c.ID), subq.c.time)
        .filter(
            Defects.c.Class == "1",
            Defects.c.IsWorstDefect == "1",
            subq.c.ID == Defects.c.BoardID,
        )
        .filter(subq.c.time.between(start_time, end_time))
        .group_by(Defects.c.Type, subq.c.time)
        .order_by(subq.c.time)
        .all()
    )
    return data


@plugmanager.register_plugin("Repair count by defect type")
def repairs_by_defect(session, metric, start_time, end_time, resolution, *_, **__):
    """Fetches data from 'get_repairs_by_defect_type' and formats data to accommodate Grafana"""
    result = []
    data = get_repairs_by_defect_type(
        session, metric, start_time, end_time, resolution, *_, **__
    )

    temp = [
        (k, list(v))
        for k, v in groupby(sorted(data, key=lambda x: x[0]), key=lambda x: x[0])
    ]

    # Sort list by defect type [(type1, count, time),(type2, count, time),
    # (type3, count, time)]"""

    for k, v in temp:
        _, ant, t = zip(*v)
        data_points = list(zip(ant, t))
        result.append({"target": k, "datapoints": data_points})

    # Formats list from [type, count, time] --> [count, time] where type is sent
    # as a target/metric to Grafana """

    return result


def get_downgradecount_by_type(session, ___, start_time, end_time, ____, *_, **__):
    """Fetches number of downgrading Defects sorted by Type"""
    Boards = DatabaseConnections.tables["Boards"]
    Defects = DatabaseConnections.tables["Defects"]
    data = (
        session.query(Defects.c.Type, func.count(Boards.c.ID))
        .join(Defects)
        .filter(Boards.c.Time.between(start_time, end_time))
        .filter(Defects.c.IsWorstDefect == "1")
        .group_by(Defects.c.Type)
        .all()
    )
    return data


@plugmanager.register_plugin("Total Defects for each board sorted by Defect type")
def defects_by_type(session, metric, start_time, end_time, resolution, *_, **__):
    """Get number of defects sorted by type and formats to Grafana"""
    data = get_downgradecount_by_type(
        session, metric, start_time, end_time, resolution, *_, **__
    )

    return [
        {
            "columns": [
                {"text": "Type", "type": "string"},
                {"text": "Antall", "type": "number"},
            ],
            "rows": data,
            "type": "table",
        }
    ]
