"""Plugin for heatmap related queries."""
from datetime import datetime

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
import seaborn as sns

from ArgosAnalysePlatform import plugmanager
from ArgosAnalysePlatform.database import DatabaseConnections


def _get_datapoints(session, no_points, start_time, end_time, *_, **__):
    Boards = DatabaseConnections.tables["Boards"]
    Defects = DatabaseConnections.tables["Defects"]

    data = (
        session.query(
            Boards.c.Length,
            Boards.c.Width,
            Defects.c.RectLeft,
            Defects.c.RectRight,
            Defects.c.RectTop,
            Defects.c.RectBottom,
        )
        .join(Boards)
        .filter(Defects.c.RectLeft != 0)
        .filter(Boards.c.Time.between(start_time, end_time))
        .order_by(Boards.c.Time)
        .limit(no_points)
        .all()
    )

    return data


@plugmanager.register_plugin("Heatmap")  # Registrerer plugin
def heatmap(session, ____, start_time, end_time, ___, extra_data, *_, **__):
    """Creates Heatmap from points taken from database

    Parameters:
        session       (Session): Creates an instance for the program to work
                        with the database
        metric        (String): Sending information to that will be used as
                        metric in Grafana(e.g. name of scanner)
        start_time    (datetime): start/from where the query will fetch data
        end_time      (datetime): end/to where the query will fetch data
        resolution    (int): control size of time-groups
        extra_data    (any): extra data to be used in program(e.g. sending
                        custom data from Grafana)
    """

    if extra_data:
        no_points = extra_data
    else:
        no_points = 1500  # standard value if not specified in Grafana

    data = _get_datapoints(session, no_points, start_time, end_time, *_, **__)

    x_max = data[0][1]
    y_max = data[0][0]
    x_array = []
    y_array = []

    for v in data:
        # x1y1
        x_array.append(v[2])
        y_array.append(v[5])
        # x2y2
        x_array.append(v[3])
        y_array.append(v[5])
        # x3y3
        x_array.append(v[3])
        y_array.append(v[4])
        # x4y4
        x_array.append(v[2])
        y_array.append(v[4])

    plt.axis([0, x_max, 0, y_max])  # limit on axis
    sns.kdeplot(
        x=x_array,
        y=y_array,
        cmap="YlOrRd",
        shade=True,
        bw_method="silverman",
    )
    # Uses Kernal density estimation(KDE) to plot density of datapoints.
    # Coordinates from DB are used to create rectangles to display defects,
    # this plot uses each corner of the rectangles as datapoints.
    # Bandwith uses silverman-method for smoothing and colormap yellow --> red.

    plt.xlabel("x-axis")
    plt.ylabel("y-axis")

    plt.savefig("ArgosAnalysePlatform/html/heat.png")
    plt.close()  # closes window

    last_update = [(datetime.now(), no_points)]

    return _format_result(last_update)


@plugmanager.register_plugin("Last defects")  # Registrerer plugin
def last_defects(session, ___, start_time, end_time, ____, extra_data, *_, **__):
    """Generate a figure over the last defects within a time range."""
    if extra_data:
        no_points = extra_data
    else:
        no_points = 1500  # standard value if not specified in Grafana

    data = _get_datapoints(session, no_points, start_time, end_time, *_, **__)
    x_max = data[0][1]
    y_max = data[0][0]

    plt.axis([0, x_max, 0, y_max])
    rectangles = plt.gca()  # get current axis

    rectangles.add_patch(
        Rectangle((0, 0), x_max, y_max, facecolor="black")
    )  # add frame to axis

    for v in data:  # creates rectangles for each defect and adds to axis
        rectangles.add_patch(
            Rectangle(
                (v[2], v[5]),
                v[3] - v[2],
                v[4] - v[5],
                facecolor="red",
            )
        )
    plt.xlabel("x")
    plt.ylabel("y")
    plt.savefig("ArgosAnalysePlatform/html/defects.png")
    plt.close()

    # current time and number of datapoints
    last_update = [(datetime.now(), no_points)]

    return _format_result(last_update)


def _format_result(last_update):
    return [
        {
            "columns": [
                {"text": "Last update", "type": "time"},
                {"text": "Number of points", "type": "number"},
            ],
            "rows": last_update,
            "type": "table",
        }
    ]
