"""Read the registry file to get classifier limits."""
import configparser
from functools import lru_cache
from math import floor

import logging
from typing import List, Dict

logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")


class ArgosRegistry:
    """Read the registry file to get classifier limits."""

    registry_parser: configparser.ConfigParser
    key_regex = r"HKEY_LOCAL_MACHINE\SOFTWARE\Argos\{}\Classifier"
    company_name: str

    @classmethod
    def from_dct(cls, dct, company_name=""):
        instance = cls()
        instance.registry_parser = dct
        instance.company_name = company_name
        instance.key_regex = ""

        return instance

    @classmethod
    def from_file(cls, file_path: str, company_name: str) -> "ArgosRegistry":
        """Read a Windows registry file and prepare it for use."""
        instance = cls()

        instance.registry_parser = configparser.ConfigParser()
        with open(file_path, "r", encoding="utf-16") as f:
            f.readline()
            instance.registry_parser.read_file(f)

        instance.key_regex = instance.key_regex.format(company_name)

        return instance

    def search_keys(self, key: str) -> List[str]:
        """Search for the specified key in the list of registry keys."""
        return [k for k in self.registry_parser.keys() if key in k]

    @lru_cache(50)
    def get(self, key: str) -> Dict[str, float]:
        """Get the limits from the chosen key.

        :param key: Name of the classifier to fetch from. Must be exact.

        :return: Dict of limits. Example {"1_2": 4}
        """
        registry_key = fr"{self.key_regex}\v{key}"

        try:
            logger.debug(f"Looking up key {registry_key}")
            section = self.registry_parser[registry_key]

            classes = {
                k.split("classlimit")[1].strip('"'): float(v.strip('"'))
                for k, v in section.items()
                if "classlimit" in k
            }

            return classes

        except KeyError:
            logger.error(
                f"Could not find key '{key}' in registry. Did you mean \n{self.search_keys(key)}"
            )
            raise

    def get_upper_limit(self, key: str, cls: int):
        """Get the upper limit of a class

        Example: For cls='0', get the registry limit classlimit1_2.

        :param key: The classifiser to look up
        :param cls: The class to get.

        """
        return floor(self.get(key)[f"{cls + 1}_{cls + 2}"])
