"""Plugin for converting other queries to XML format."""
import xml.etree.ElementTree as ET
from datetime import datetime
from typing import Dict, Union, Iterable

from sqlalchemy import func
from sqlalchemy.orm import Session

from ArgosAnalysePlatform import DatabaseConnections, plugmanager
from ArgosAnalysePlatform.plugins.boards import group_by_time


def xml_converter(dataset: list, column_names: list):
    """Converts SQL Alchemy query results to XML

    :param dataset:         List of tuples containing query data
    :param column_names:    List of strings that names columns in chronological order.

    :return:                Iterable XML element
    """

    # If no data is retrieved, return named columns with empty values
    if not dataset:
        row = ET.Element("Row")
        for part in dataset:
            ET.SubElement(row, part)

        yield row

    for item in dataset:
        # Set row child-node for each individual result in dataset
        row = ET.Element("Row")

        for i, part in enumerate(item):
            # Set each part of every result as individual value child-nodes
            ET.SubElement(row, column_names[i]).text = str(part)

        yield row


@plugmanager.register_plugin("ClassOverview", category="report_builder")
def xml_get_class_overview(
    session: Session,
    _,
    start_time: datetime,
    end_time: datetime,
    resolution: int,
    extra_data: Dict[str, Union[str, int]],
) -> Iterable[ET.Element]:
    """Gets count of each class between two points in time

    :param session:     Current database session.
    :param start_time:  Start of query time limit.
    :param end_time:    End of query time limit.
    :param resolution:  The size of each time-group (ms).
    :param extra_data:  Any extra data, in this case sorting parameter.

    :return:            Iterable of XML Elements containing query data
    """
    Boards = DatabaseConnections.tables["Boards"]
    subq = group_by_time(session, resolution)
    column_names = ["Class", "Count", "Time"]

    data = (
        session.query(subq.c.Class, func.count(subq.c.ID), subq.c.time)
        .join(Boards, Boards.c.ID == subq.c.ID)
        .filter(subq.c.time.between(start_time, end_time))
    )

    # If group-by column is included in call, filter data by it
    if "gbcolumn_name" in extra_data:
        data = data.filter(
            getattr(Boards.c, extra_data["gbcolumn_name"])
            == extra_data["gbcolumn_value"]
        )

    data = data.group_by(subq.c.Class, subq.c.time)

    return xml_converter(data.all(), column_names)


@plugmanager.register_plugin("ProductionInfo", category="report_builder")
def xml_get_production_info(
    session: Session,
    _,
    start_time: datetime,
    end_time: datetime,
    __,
    ___,
) -> Iterable[ET.Element]:
    """Gets information about production between two points in time

    :param session:     Current database session.
    :param start_time:  Start of query time limit.
    :param end_time:    End of query time limit.

    :return:            Iterable of XML Elements containing query data.
    """
    Boards = DatabaseConnections.tables["Boards"]
    column_names = [
        "Time",
        "Product",
        "Series",
        "Shift",
        "Length",
        "Width",
        "Thickness",
        "ClassName",
    ]

    data = (
        session.query(
            Boards.c.Time,
            Boards.c.Product,
            Boards.c.Series,
            Boards.c.Shift,
            Boards.c.Length,
            Boards.c.Width,
            Boards.c.Thickness,
            Boards.c.ClassName,
        )
        .filter(Boards.c.Time.between(start_time, end_time))
        .all()
    )

    return xml_converter(data, column_names)


@plugmanager.register_plugin("Downgrades", category="report_builder")
def xml_get_downgrades(
    session: Session,
    _,
    start_time: datetime,
    end_time: datetime,
    __,
    extra_data: Dict[str, Union[str, int]],
) -> Iterable[ET.Element]:
    """Gets downgrade type, corresponding count and initial class

    :param session:     Current database session.
    :param start_time:  Start of query time limit.
    :param end_time:    End of query time limit.
    :param extra_data:  Any extra data, in this case sorting parameter.

    :return:            Iterable of XML Elements containing query data
    """

    Boards = DatabaseConnections.tables["Boards"]
    Defects = DatabaseConnections.tables["Defects"]
    column_names = ["DefectType", "Count", "Class"]

    data = (
        session.query(Defects.c.Type, func.count(Boards.c.ID), Defects.c.Class)
        .join(Defects)
        .filter(Boards.c.Time.between(start_time, end_time))
        .filter(Defects.c.IsWorstDefect == "1")
    )

    # If group-by column is included in call, filter data by it
    if "gbcolumn_name" in extra_data:
        data = data.filter(
            getattr(Boards.c, extra_data["gbcolumn_name"])
            == extra_data["gbcolumn_value"]
        )

    data = data.group_by(Defects.c.Type, Defects.c.Class)

    return xml_converter(data.all(), column_names)
