from pathlib import Path

import random
import tempfile
from fastapi import Request
from fastapi.testclient import TestClient
import pytest
from sqlalchemy import create_engine, Column, Integer, DateTime, String
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

from ArgosAnalysePlatform import global_dependencies, APIApp
from ArgosAnalysePlatform.database import DatabaseConnections


class MockDatabaseConnections:
    def __init__(self, names):
        self.names = names

    def sessioniter(self, *_, **__):
        yield from [(name, "session_placeholder") for name in self.names]


def override_get_db(request: Request):
    request.state.db = MockDatabaseConnections(["scanner_name"])


def _get_injected_api_app(injected_db_callable):
    app = APIApp(
        load_from_settings={"reflect": False, "debug": False, "use_database": False}
    )
    app.dependency_overrides[global_dependencies["get_db"]] = injected_db_callable
    return TestClient(app)


@pytest.fixture(name="client", autouse=True)
def init_tests():
    return _get_injected_api_app(override_get_db)


@pytest.fixture(name="memory_database")
def create_in_memory_database():
    tmp_file = tempfile.NamedTemporaryFile(suffix=".db").name
    engine = create_engine(f"sqlite:///{tmp_file}")
    smaker = sessionmaker(bind=engine)

    database = DatabaseConnections({})
    database.connections = {"memory": {"engine": engine, "sessionmaker": smaker}}

    Base = declarative_base()

    class MockedBoards(Base):
        __tablename__ = "Boards"
        ID = Column(Integer, primary_key=True)
        BoardID = Column(Integer)
        ClassName = Column(String)
        Class = Column(Integer)
        Time = Column(DateTime)

        def __repr__(self):
            return f"MockedBoards(ID={self.ID})"

    class MockedDefects(Base):
        __tablename__ = "Defects"
        ID = Column(Integer, primary_key=True)
        Type = Column(String)
        SubType = Column(String)
        ArgosName = Column(String)

        def __repr__(self):
            return f"MockedDefects(ID={self.ID})"

    DatabaseConnections.tables = {}
    DatabaseConnections.tables["Boards"] = MockedBoards.__table__
    DatabaseConnections.tables["Defects"] = MockedDefects.__table__

    Base.metadata.create_all(engine)

    def override_get_db_memory(request: Request):
        request.state.db = database

    return database, _get_injected_api_app(override_get_db_memory)
